/*
* This file is part of Firestorm NIDS
* Copyright (c) 2001 Gianni Tedesco
*
*    Firestorm is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    Firestorm is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with Firestorm; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include <loader.h>
#include <s_globals.h>
#include <s_util.h>
#include <cleanup.h>
#include <matcher.h>
#include <rules.h>

p_plugin_list	ld_list=NULL;
const plugin_in_t g_export={
	size:		sizeof(plugin_in_t),
	print_out:	print_out,
	print_raw:	print_raw,
	file_out:	file_out,
	file_raw:	file_raw,
	malloc:		malloc,
	strdup:		strdup,
	free:		free,
	add_captype:	capdev_add_type,
	add_demux:	demux_add_handler,
	add_matcher:	matcher_add,
	add_target:	target_add,
	add_rule:	matcher_add_rule,
	add_criteria:	matcher_add_criteria_to_rule,
	rule_cleanup:	matcher_rule_cleanup,
	set_target:	matcher_set_rule_target,
	add_parser:	rules_add_parser,
	list_append:	list_append,
	list_remove:	list_remove,
	get_var:	plugin_get_var,
};

int loader_init(void)
{
	return 1;
}

void loader_load_plugin(char *path)
{
	p_plugin_list	plugin=NULL;
	plugin_in_t	export;
	plugin_out_t	import;
	
	if ( globals.num_plugins >= MAX_PLUGINS ) {
		print_out(PRINT_ERROR "MAX_PLUGINS reached (%i)" CRLF, 
			MAX_PLUGINS);
		return;
	}

	if ( path == NULL ) return;

	/* Allocate the mem for this plugin in the list */	
	if ( !(plugin=globals.malloc(sizeof(plugin_list_t))) ) {
		return;
	}else{
		memset(plugin, 0, sizeof(plugin_list_t));
	}

	/* Open the plugin file */
	if ( !(plugin->lib=dlopen(path, PLUGIN_MODE )) ) {
		print_out(PRINT_ERROR "Couldn't open plugin (%s)" CRLF,path);
		globals.free(plugin);
		return;
	}

	/* Resolve the API symbols */
	plugin->init   = dlsym(plugin->lib, PLUGIN_INIT);
	plugin->unload = dlsym(plugin->lib, PLUGIN_UNLOAD);

	if ( (plugin->init==NULL) || (plugin->unload==NULL) ) {
		print_out(PRINT_ERROR 
			"Couldn't resolve plugin symbols (%s)" CRLF,path);
		dlclose(plugin->lib);
		globals.free(plugin);
		return;
	}
	
	/* Setup stuff to export to plugin */
	memset(&import, 0, sizeof(import));
	memcpy(&export, &g_export, sizeof(export));
	template_fill_struct(&export.templates);
	
	/* Jump to the plugin entry point */
	switch ( plugin->init(&export, &import) ) {
		case PLUGIN_ERR_OK:
			print_out("plugin: %s v%i.%i" 
				" - %s"	CRLF, import.name, 
				import.ver_major, import.ver_minor,
				import.desc);
			break;
	
		case PLUGIN_ERR_NOMEM:
			print_out(PRINT_ERROR "Plugin init failed (%s): "
				"Out of memory" CRLF, path);
			dlclose(plugin->lib);
			globals.free(plugin);
			return;

		case PLUGIN_ERR_PARAM:
			print_out(PRINT_ERROR "Plugin init failed (%s): "
				"Bad parameter" CRLF, path);
			dlclose(plugin->lib);
			globals.free(plugin);
			return;

		case PLUGIN_ERR_SIZE:
			print_out(PRINT_ERROR "Plugin init failed (%s): "
				"Wrong version, please recompile" CRLF, path);
			dlclose(plugin->lib);
			globals.free(plugin);
			return;
			
		case PLUGIN_ERR_FAIL:
		default:
			print_out(PRINT_ERROR "Plugin init failed (%s): " 
				"Unspecified reason" CRLF, path);
			dlclose(plugin->lib);
			globals.free(plugin);
			return;
	}
	
	/* grab what we need */
	plugin->name=import.name;
	plugin->desc=import.desc;
	plugin->ver_minor=import.ver_minor;
	plugin->ver_major=import.ver_major;
	
	/* prepend to the list */
	plugin->next=ld_list;
	ld_list=plugin;
	globals.num_plugins++;
}

/* Remember a trailing slash on the dir!  */
int loader_load_dir(char *plugin_dir, int recurse_depth)
{
	DIR *dir=NULL;
	struct dirent *file=NULL;
	struct stat sfile;
	char fullpath[1024]; /* Bigger than PATH_MAX anywhere */

	if ( !(dir=opendir(plugin_dir)) )
		return 0;

	while(dir) {
		if ((file=readdir(dir)) == NULL) break;

		memset(&fullpath, 0, 256);
		strncpy((char *)&fullpath, plugin_dir, 255 );
		strncat((char *)&fullpath, "/", 2);
		strncat((char *)&fullpath, file->d_name, 253 );

		if ( stat(fullpath, &sfile) ) continue;
		if ( (sfile.st_mode & S_IFREG) ) {
			loader_load_plugin((char *)&fullpath);
		} else if ( (sfile.st_mode & S_IFDIR) && recurse_depth ) {
			loader_load_dir((char *)&fullpath, recurse_depth-1);
		}
	}closedir(dir);

	return 1;
}

void loader_cleanup(int code, void *priv)
{
	p_plugin_list	tmp=NULL;
	p_plugin_list   foo=NULL;
	
	print_raw("Unloading all plugins" CRLF);

	for (tmp=ld_list; tmp; tmp=tmp->next) {
		if ( tmp->unload() != PLUGIN_ERR_OK ) {
			print_out(PRINT_WARNING "%s plugin reported "
				"problems unloading" CRLF, tmp->name);
		}
		dlclose(tmp->lib);
	}

	for (tmp=ld_list; tmp;) {
		foo=tmp;
		tmp=tmp->next;
		globals.free(foo);
		foo=NULL;
	}
}
