/*
* This file is part of Firestorm NIDS
* Copyright (c) 2001 Gianni Tedesco
*
*    Firestorm is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    Firestorm is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with Firestorm; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <matcher.h>
#include <s_util.h>
#include <cleanup.h>
#include <targets.h>

#define MAX_HDR 5

/* List of registered matchers   */
p_matcher	matches=NULL;

/* List of all rules */
p_rule		rules;

/* All unique rule elements */
p_relement	*all_priv=NULL;
u_int32_t	num_priv=0;

/* chosen matchers */
p_matcher 	*all_m=NULL;
u_int32_t	num_m=0;

/* Rule header chain */
p_cheader	hdr=NULL;
u_int32_t	hdr_count;

p_relement matcher_copyrel(p_relement c)
{
	p_relement ret;

	if ( !(ret=malloc(sizeof(relement_t))) ) {
		return NULL;
	}

	memcpy(ret, c, sizeof(relement_t));
	return ret;
}

p_cheader matcher_mkheader(p_rule r, p_matcher *m, u_int32_t count)
{
	p_cheader ret;
	p_relement c, thisc, tmp;
	u_int32_t i;

	if ( !(ret=calloc(1, sizeof(p_matcher))) )
		return NULL;
	
	for(i=0; i<count; i++) {
		for(c=r->criteria; c;) {
			thisc=c;
			c=(p_relement)c->l.next;
			
			if ( thisc->match == m[i] ) {

				if ( !(tmp=matcher_copyrel(thisc)) ) {
					free(ret);
					return 0;
				}

				if (thisc->l.prev) {
					list_remove(thisc);
				}else{
					((p_relement)thisc->l.next)->l.prev=thisc;
					r->criteria=(p_relement)thisc->l.next;
				}
				ret->criteria=list_append(tmp, ret->criteria);
				break;
			}
			if ( r->criteria==NULL ) return ret;
		}
	}
	
	return ret;
}

/* Is a exactly equal to b */
int matcher_comparehdr(p_relement a, p_relement b)
{
	u_int32_t a_count=0, b_count=0;
	u_int32_t match=0;
	p_relement tmp_a, tmp_b;

	if ( !a || !b ) return 0;
	
	/* make sure both sides have the same number of elements */
	for(tmp_a=a; tmp_a; tmp_a=(p_relement)tmp_a->l.next) a_count++;
	for(tmp_b=b; tmp_b; tmp_b=(p_relement)tmp_b->l.next) b_count++;
	if ( a_count != b_count ) return 0;

	/* check everything in a is also in b */
	for(tmp_a=a; tmp_a; tmp_a=(p_relement)tmp_a->l.next) {
		for(tmp_b=b; tmp_b; tmp_b=(p_relement)tmp_b->l.next) {
			if ( tmp_a->match==tmp_b->match &&
				tmp_a->negate==tmp_b->negate &&
				tmp_a->priv==tmp_b->priv) {
				match++;
				break;
			}
		}
	}
	
	return (match==a_count) ? 1 : 0;
}

int matcher_rule_to_chain(p_rule r, p_rule *ptr)
{
	p_cheader h, tmp;
	int m=0;

	if ( !(h=matcher_mkheader(r, all_m, num_m)) )
		return 0;
	
	for(tmp=hdr; tmp; tmp=h->next) {
		if ( matcher_comparehdr(h->criteria, hdr->criteria) ) {
			list_delete(r, ptr);
			list_append(r, hdr->rules);
			free(h);
			return 1;
		}
	}

	/* Add a new header to the chain */
	list_delete(r, ptr);
	h->rules=r;
	h->next=hdr;
	hdr=h;
	hdr_count++;

	return 1;
}

int matcher_dopriv(void)
{
	p_rule r;
	p_relement c;
	u_int32_t count=0;
	u_int32_t i,u;
	
	for(r=rules; r; r=(p_rule)r->l.next) {
		for(c=r->criteria; c; c=(p_relement)c->l.next) {
			count++;
		}
	}

	if ( !(all_priv=calloc(count, sizeof(p_relement))) ) return 0;
	for(r=rules; r; r=(p_rule)r->l.next) {
		for(c=r->criteria; c; c=(p_relement)c->l.next) {
			u=1;
			
			/* Try find an identical criteria */
			for(i=0; i<num_priv; i++) {
				p_relement a=all_priv[i];
				if ( a->match==c->match && c->match->compare(
					a->priv, c->priv)==CMP_IDENTICAL ) {
					c->match->cleanup(c->priv);
					c->priv=a->priv;
					a->usage++;
					u=0;
					break;
				}
			}

			/* It's unique */
			if ( u ) {
				all_priv[num_priv]=c;
				num_priv++;
			}
		}
	}

	/* shrink the memory we got */
	realloc(all_priv, sizeof(p_relement)*num_priv);
	
	print_out("RULESET: all/unique/redundant: %u/%u/%u" CRLF,
		count, num_priv, count-num_priv);

	return 1;
}

int m_compare(const matcher_t **a, const matcher_t **b)
{
	if ( (*a)->common > (*b)->common ) return -1;
	if ( (*a)->common < (*b)->common ) return 1;
	return 0;
}

int matcher_compile(void)
{
	u_int32_t i;
	p_matcher m;
	p_rule r,thisr;

	print_out("RULESET: %u rule files loaded, "
		"%u rules all together." CRLF,
		globals.num_rulefiles,
		globals.num_rules);
	
	/* 1. Make list of unique criteria */
	if ( !matcher_dopriv() ) return 0;
	
	/* 2. Use list to mark most common matchers */
	for(i=0; i<num_priv; i++) {
		if ( all_priv[i]->usage > all_priv[i]->match->common ) {
			all_priv[i]->match->common=all_priv[i]->usage;
		}
	}
	
	/* 3. Make sorted list of most common matchers */
	if ( !(all_m=calloc(globals.num_matchers, sizeof(p_matcher))) )
		return 0;
	
	for(m=matches, i=0; m; m=m->next, i++)
		all_m[i]=m;
	
	qsort(all_m, globals.num_matchers, sizeof(p_matcher), m_compare);
	
	/* 4. Pick the top n matchers */
	num_m=MAX_HDR > globals.num_matchers ? globals.num_matchers : MAX_HDR;
	realloc(all_m, num_m*sizeof(p_matcher));
	
	print_raw("RULESET: ");
	for(i=0; i<num_m; i++) {
		print_out("%s -> ", all_m[i]->name, all_m[i]->common);
	}print_raw("NULL" CRLF);
	
	/* 5. Build header chain from rules */
	for(r=rules; r; ) {
		thisr=r;
		r=(p_rule)r->l.next;
		matcher_rule_to_chain(thisr, &rules);
		if ( !rules ) break;
	}

	print_out("RULESET: %u header nodes" CRLF, hdr_count);

	/* cleanup */
	free(all_m);
	return 1;
}


int matcher_add_rule(p_rule r)
{
	if ( r == NULL ) return 0;

	/* FIXME: validate rule better */
	if ( r->criteria == NULL ) return 0;

	rules=list_append(r, rules);
	globals.num_rules++;

	return 1;
}

int matcher_set_rule_target(p_rule r, char *target, char *val)
{
	p_target t;
	if ( !r || !target ) return 0;
	if ( !(t=target_get(target)) ) return 0;
	if ( !(r->t_priv=t->validate(val)) ) return 0;
	r->target=t;
	return 1;
}

/* Add a criteria to a rule */
int matcher_add_criteria_to_rule(
	p_rule r, 
	char *crit,
	char *val,
	p_kv opts,
	int negate)
{
	p_matcher matcher=NULL;
	p_relement rel=NULL;
	
	if ( !r || !crit || !val ) return 0;

	if ( (matcher=matcher_get(crit)) == NULL ) {
		print_out(PRINT_ERROR "Unable to add rule, no such "
			"matcher (%s)" CRLF, crit);
		return 0;
	}

	if ( (rel=globals.malloc(sizeof(relement_t))) == NULL  ) {
		return 0;
	}

	/* Validate this rule criteria */
	rel->match = matcher;
	if ( (rel->priv=rel->match->validate(val, opts)) == NULL ) {
		globals.free(rel);
		return 0;
	}

	rel->negate = (negate) ? 1 : 0;
	
	/* Now add the criteria to the rule */
	rel->usage=0;
	r->criteria = list_append(rel, r->criteria);
	return 1;
}

/* 
* Cleans up a rule object (MUST have been
* safely initialised with SAFE_INIT_RULE_PTR() 
* before calling this
*/
void matcher_rule_cleanup(p_rule r)
{
	p_relement r_foo, r_bar;

	if ( r->target && r->t_priv )
		r->target->cleanup(r->t_priv);
	
	for (r_foo=r->criteria; r_foo;) {
		r_bar=r_foo;
		r_foo=(p_relement)r_foo->l.next;
		globals.free(r_bar);
	}

	globals.free(r);
}

void matcher_cleanup(int code, void *priv)
{
	p_rule foo,bar;
	p_matcher m_foo, m_bar;
	p_relement p;
	u_int32_t i;
	
	/* TODO: Fix... */
	print_raw("Removing all rules" CRLF);
	for (foo=rules; foo;) {
		bar=foo;
		foo=(p_rule)foo->l.next;
		matcher_rule_cleanup(bar);
	}

	print_raw("Removing matcher private data" CRLF);
	for(i=0; i<num_priv; i++) {
		all_priv[i]->match->cleanup(all_priv[i]->priv);
		all_priv[i]->priv=NULL;
	}

	print_raw("Removing all matchers" CRLF);
	for (m_foo=matches; m_foo;) {
		m_bar=m_foo;
		m_foo=m_foo->next;
		globals.free(m_bar);
	}
}

int matcher_init(void)
{
	cleanup_add_handler(matcher_cleanup, NULL);
	return 1;
}

/* Get a matcher by its name  */
p_matcher matcher_get(char *n)
{
	p_matcher tmp;

	for ( tmp=matches; tmp; tmp=tmp->next ) {
		if ( strcasecmp(tmp->name, n) == 0 ) {
			return tmp;
		}
	}

	return NULL;
}

/* Plugins call this to add a new matcher */
int matcher_add ( char *n, proc_match_validate v, 
		proc_match_cleanup cl,
		proc_match_match match,
		proc_match_compare cmp)
{
	p_matcher foo;

	if ( n == NULL ) return 0;
	if ( v == NULL ) return 0;
	if ( cl == NULL ) return 0;
	if ( cmp == NULL ) return 0;
	
	if ( matcher_get(n) ) {
		print_out(PRINT_WARNING 
			"Tried to load %s matcher twice" CRLF, n);
		return 0;
	}
	
	if ( !(foo=globals.malloc(sizeof(matcher_t))) ) {
		return 0;
	}

	foo->name=n;
	foo->validate=v;
	foo->cleanup=cl;
	foo->compare=cmp;
	foo->match=match;
	foo->common=0;
	
	foo->next=matches;
	matches=foo;
	globals.num_matchers++;

	return 1;
}

/* Actually match a packet against all our rules */
static int inline matcher_begin_rulechain( p_packet p, p_rule rulechain )
{	
	int m=1;
	p_rule r=NULL;
	p_relement c=NULL;
#if 0
	p_target def;

	if ( rules == NULL ) {
		if ( (def=target_get("ALERT")) )
			def->target(p, "PACKET RECIEVED", rules);
		else
			print_out("ALERT target not loaded, and no rules");
		return;
	}
#endif

	for (r=rulechain; r; r=(p_rule)r->l.next) {
		m=1;
		
		for (c=r->criteria; c; c=(p_relement)c->l.next) {
			globals.num_ops++;
			if ( !(c->match->match(c->priv, p, c->negate)) ){
				m=0;
				break;
			}
		}
		
		if (m) {
			/* Send the packet to the rule target */
			r->target->target(p, r->t_priv, r);
			return 1;
		}
	}

	return 0;
}

void inline matcher_begin( p_packet p )
{
	p_cheader h;
	p_relement c;
	int m;

	for(h=hdr; h; h=h->next) {
		m=1;
		for (c=h->criteria; c; c=(p_relement)c->l.next) {
			globals.num_ops++;
			if ( !(c->match->match(c->priv, p, c->negate)) ){
				m=0;
				break;
			}
		}

		if ( m ) {
			/* TODO: MTF heuristic */
			if ( matcher_begin_rulechain( p, h->rules ) ) {
				return;
			}
		}
	}
}
