/*
* This file is part of Firestorm NIDS
* Copyright (c) 2001 Gianni Tedesco
*
*    Firestorm is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    Firestorm is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with Firestorm; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <s_globals.h>
#include <s_util.h>
#include <sensor.h>
#include <cleanup.h>
#include <targets.h>
#include <loader.h>
#include <capdev.h>
#include <demux.h>

/* Our one and only global structure */
struct sensor_globals_t globals;

/* Daemonize ourself and put ourselves in to
* a known state. Under UNIX processes inherit
* all kinds of crap, we need to obliterate it all */
int daemon_boogie(const char *lfn) 
{
	int lfd;
	int fd=0,fds;

	/* Close all file descriptors, we could
	* have inherited fds from parent processes etc.. */
	fds=sysconf(_SC_OPEN_MAX);
	while (fd<fds) close(fd++);
	
	/* Set our umask, this is inherited */
	umask(0);
	
	/* First fork off in to 
	* the background */
	switch( fork() ) {
		case 0: break;
		case -1: return 0;
		default: _exit(0);
	}

	/* Create a new session and become
	* session group leader */
	if ( setsid() < 0 ) {
		return 0;
	}

	/* Fork once again, we now loose
	* control of the session and can
	* never regain a tty ever again */
	switch( fork() ) {
		case 0: break;
		case -1: return 0;
		default: _exit(0);
	}

	/* Open our log file */
	if ( (lfd=open(lfn, O_WRONLY | O_CREAT | O_TRUNC, 00600)) <0 )
		return 0;

	/* dup stdout/stderr to the log file */
	if ( dup2(lfd, 1) < 0 ) return 0;
	if ( dup2(lfd, 2) < 0 ) return 0;
	
	return 1;
}

const char *plugin_get_var(const char *name)
{
	p_kv t;

	if ( !name ) return NULL;
	
	for(t=globals.plugin_vars; t; t=(p_kv)t->l.next) {
		if ( !strcasecmp(name, t->key) ) {
			return t->val;
		}
	}
	
	return NULL;
}

inline void capture_proc(void *parm)
{
	p_capdev c;
	p_packet p;
	u_int32_t a=0;
	
	c=(p_capdev)parm;
	c->state=CAP_STATE_CAPTURE;
	
	for(;;){
		/* Poll the capture device */
		if ( (p=c->capture->poll(c->priv)) == NULL ) break;
		p->capdev=c;
		
		/* Demultiplex... */
		p->top_layer=demux(p);
		
		/* Match */
		matcher_begin(p);

		/* Per capdev packet count */
		a++;
	}

	print_out("Ending %s capture [%s]: Caught %u packets." CRLF, 
		c->capture->name, c->interface, a);

	c->capture->end(c->priv);
	c->state=CAP_STATE_NULL;
	
	return;
}

int main ( int argc, char **argv ) 
{
	/* Curse me for this needless waste of stack space in main() */
	char *lfn=NULL;

	/* Set timezone from locale */
	tzset();
	
	memset(&globals, 0, sizeof(globals));
	globals.malloc=malloc;
	globals.free=free;
	globals.fs_strdup=(proc_fs_strdup)strdup;

	/* print the fersion string on to the console */
	print_raw(VERSION_STRING);

	if ( argc < 2 || argc > 3) {
		print_out("Usage: %s <configfile> [<logfile>]" CRLF, argv[0]);
		return EXIT_ERR;
	}else{
		globals.config_fn=argv[1];
	}
	
	/* Can't go nowhere with no cleanup subsystem */
	if ( !cleanup_init() ) return EXIT_ERR;

#if 1
	/* Do the daemon boogie */
	lfn=(argc == 3) ? argv[2] : "firestorm.log";
	if ( !daemon_boogie(lfn) ) {
		cleanup(EXIT_ERR);
	}
#endif

	/* Also print version info to file */
	print_raw(VERSION_STRING CRLF);
	
	/* Load capture subsystem */
	if ( !capdev_init() ) cleanup(EXIT_ERR);

	/* Load demultiplexing subsystem */
	if ( !demux_init() ) cleanup(EXIT_ERR);
	
	/* Load packet target subsystem */
	if ( !target_init() ) cleanup(EXIT_ERR);
	
	/* Load packet matcher subsystem  */
	if ( !matcher_init() ) cleanup(EXIT_ERR);
	
	/* Load ruleset file subsytem  */
	if ( !rules_init() ) cleanup(EXIT_ERR);

	/* Load and run our config */
	if ( !conf_load_config(globals.config_fn) ) {
		conf_cleanup_temp();
		cleanup(EXIT_ERR);
	}
	if ( !conf_go() ) {
		conf_cleanup_temp();
		cleanup(EXIT_ERR);
	}conf_cleanup_temp();

	/* Print a quick summary */
	print_out("PLUGINS: %u plugins loaded: "
		"%u matchers, "
		"%u targets, %u captypes." CRLF,
		globals.num_plugins,
		globals.num_matchers,
		globals.num_targets,
		globals.num_captypes);
	
	matcher_compile();
	
	/* Multiple captures are broken */
	print_out("%s capture engine: started [%s]" CRLF,
		capdevs->capture->name,
		(capdevs->interface) ? capdevs->interface : "any");

	gettimeofday(&globals.start_time, NULL);
	capture_proc(capdevs);
	gettimeofday(&globals.end_time, NULL);

	print_out(" STARTED AT: %u.%u" CRLF,
		globals.start_time.tv_sec,
		globals.start_time.tv_usec);
	print_out("FINISHED AT: %u.%u" CRLF,
		globals.end_time.tv_sec,
		globals.end_time.tv_usec);
	cleanup(EXIT_OK);
}
